import numpy as np
from ..easyviewer.base import SEGMENTATION_LAYER_TYPES


def layer_names(state):
    """Get all layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l['name'] for l in state['layers']]


def image_layers(state):
    """Get all image layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l['name'] for l in state['layers'] if l['type'] == 'image']


def segmentation_layers(state):
    """Get all segmentation layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l['name'] for l in state['layers'] if l['type'] in SEGMENTATION_LAYER_TYPES]


def annotation_layers(state):
    """Get all annotation layer names in the state

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict

    Returns
    -------
    names : list
        List of layer names
    """
    return [l['name'] for l in state['layers'] if l['type'] == 'annotation']


def tag_dictionary(state, layer_name):
    """Get the tag id to string dictionary for a layer

    Parameters
    ----------
    state : dict

    layer_name : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    l = get_layer(state, layer_name)
    taginfo = l.get('annotationTags', [])
    tags = {}
    for t in taginfo:
        tags[int(t['id'])] = t['label']
    return tags


def get_layer(state, layer_name):
    """Gets the contents of the layer based on the layer name.

    Parameters
    ----------
    state : dict
        Neuroglancer state as a JSON dict
    layer_name : str
        Name of layer

    Returns
    -------
    layer : dict
        Layer data contents
    """
    layer_ind = layer_names(state).index(layer_name)
    return state['layers'][layer_ind]


def view_settings(state):
    """Get all data about the view state in neuroglancer: position,
    image zoom, orientation and zoom of the 3d view, and voxel size.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict

    Returns
    -------
    view : dict
        Dictionary with keys: position, zoomFactor, perspectiveOrientation,
        perspectiveZoom, and voxelSize
    """
    view = {}
    view['position'] = state['navigation']['pose']['position']['voxelCoordinates']
    view['zoomFactor'] = state['navigation'].get('zoomFactor', None)
    view['perspectiveOrientation'] = state.get('perspectiveOrientation', None)
    view['perspectiveZoom'] = state.get('perspectiveZoom', None)
    view['voxelSize'] = state['navigation']['pose']['position']['voxelSize']
    return view


def _get_type_annotations(state, layer_name, type):
    l = get_layer(state, layer_name)
    annos = l.get('annotations', [])
    return [anno for anno in annos if anno['type'] == type]


def _get_point_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, 'point')


def _get_sphere_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, 'ellipsoid')


def _get_line_annotations(state, layer_name):
    return _get_type_annotations(state, layer_name, 'line')


def _extract_point_data(annos):
    return [[anno.get('point') for anno in annos]]


def _extract_sphere_data(annos):
    pt = [anno.get('center') for anno in annos]
    rad = [anno.get('radii') for anno in annos]
    return [pt, rad]


def _extract_line_data(annos):
    ptA = [anno.get('pointA') for anno in annos]
    ptB = [anno.get('pointB') for anno in annos]
    return [ptA, ptB]


_get_map = {
    'point': _get_point_annotations,
    'line': _get_line_annotations,
    'sphere': _get_sphere_annotations,
}

_extraction_map = {
    'point': _extract_point_data,
    'line': _extract_line_data,
    'sphere': _extract_sphere_data
}


def _generic_annotations(state, layer_name, description, linked_segmentations, tags, type):
    annos = _get_map[type](state, layer_name)
    out = _extraction_map[type](annos)
    if description:
        desc = [anno.get('description', None) for anno in annos]
        out.append(desc)
    if linked_segmentations:
        linked_seg = [[np.uint64(x) for x in anno.get(
            'segments', [])] for anno in annos]
        out.append(linked_seg)
    if tags:
        tag_list = [anno.get('tagIds', []) for anno in annos]
        out.append(tag_list)
    if len(out) == 1:
        return out[0]
    else:
        return out


def point_annotations(state, layer_name, description=False, linked_segmentations=False, tags=False):
    """Get all point annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points : list
        List of N 3-element points (as list)
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    """
    return _generic_annotations(state, layer_name, description, linked_segmentations, tags, 'point')


def line_annotations(state, layer_name, description=False, linked_segmentations=False, tags=False):
    """Get all line annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points_A : list
        List of N 3-element points (as list) of the first point in each line.
    anno_points_B : list
        List of N 3-element points (as list) of the second point in each line.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    """
    return _generic_annotations(state, layer_name, description, linked_segmentations, tags, 'line')


def sphere_annotations(state, layer_name, description=False, linked_segmentations=False, tags=False):
    """Get all sphere annotation points and other info from a layer.

    Parameters
    ----------
    state : dict
        Neuroglancer state as JSON dict
    layer_name : str
        Layer name
    description : bool, optional
        If True, also returns descriptions as well. By default False
    linked_segmentations : bool, optional
        If True, also returns list of linked segmentations, by default False
    tags : bool, optional
        If True, also returns list of tags, by default False

    Returns
    -------
    anno_points : list
        List of N 3-element center points (as list)
    radius_points : list
        List of N 3-element radii for each axis of the ellipsoid.
    anno_descriptions : list
        List of N strings (or None), only returned if description=True.
    anno_linked_segmentations : list
        List of N lists of object ids. Only returned if linked_segmentations=True.
    anno_tags : list
        List of N lists of tag ids. Only returned if tags=True.
    """
    return _generic_annotations(state, layer_name, description, linked_segmentations, tags, 'sphere')



from nglui import EasyViewer, annotation
from nglui.easyviewer.utils import default_neuroglancer_base
import pandas as pd
import numpy as np
from collections.abc import Collection
from IPython.display import HTML
from .utils import bucket_of_values

DEFAULT_IMAGE_LAYER = 'img'
DEFAULT_SEG_LAYER = 'seg'
DEFAULT_ANNO_LAYER = 'anno'

DEFAULT_VIEW_KWS = {'layout': 'xy-3d',
                    'zoom_image': 2,
                    'show_slices': False,
                    'zoom_3d': 2000,
                    }

DEFAULT_SEGMENTATION_VIEW_KWS = {'alpha_selected': 0.3,
                                 'alpha_3d': 1,
                                 'alpha_unselected': 0}


class LayerConfigBase(object):
    """Base class for configuring layers

    Parameters
    ----------
    name : str
        Layer name for reference and display
    type : str
        Layer type. Usually handled by the subclass
    source : str
        Datasource for the layer
    color : str
        Hex string (with starting hash).
    active : bool,
        If True, becomes a selected layer.
    """

    def __init__(self,
                 name,
                 type,
                 source,
                 color,
                 active,
                 ):
        self._config = dict(type=type,
                            name=name,
                            source=source,
                            color=color,
                            active=active,
                            )

    @property
    def type(self):
        return self._config.get('type', None)

    @property
    def name(self):
        return self._config.get('name', None)

    @name.setter
    def name(self, n):
        self._config['name'] = n

    @property
    def source(self):
        return self._config.get('source', None)

    @source.setter
    def source(self, s):
        self._config['source'] = s

    @property
    def color(self):
        return self._config.get('color', None)

    @color.setter
    def color(self, c):
        self._config['color'] = c

    @property
    def active(self):
        return self._config.get('active', False)

    def _render_layer(self, viewer, data):
        """ Applies rendering rules
        """
        self._specific_rendering(viewer, data)
        if self.active:
            viewer.set_selected_layer(self.name)
        return viewer

    def _specific_rendering(self, viewer, data):
        """ Subclasses implement specific rendering rules
        """
        pass


class ImageLayerConfig(LayerConfigBase):
    """Image layer configuration class.

    This provides the rules for setting up an image layer in neuroglancer.

    Parameters
    ----------
    source : str
        Cloudpath to an image source
    name : str, optional
        Name of the image layer. By default, 'img'.
    active : bool, optional
        If True, makes the layer active in neuroglancer. Default is False.
    contrast_controls : bool, optional
        If True, gives the layer a user-controllable brightness and contrast shader. Default is False.
    black : float, optional
        If contrast_controls is True, sets the default black level. Default is 0.0.
    white : float, optional
        If contrast_controls is True, sets the default white level. Default is 1.0.
    """

    def __init__(self,
                 source,
                 name=DEFAULT_IMAGE_LAYER,
                 active=False,
                 contrast_controls=False,
                 black=0.0,
                 white=1.0,
                 ):
        super(ImageLayerConfig, self).__init__(
            name=name, type='image', source=source, color=None, active=active)
        self._contrast_controls = contrast_controls
        self._black = black
        self._white = white

    def _specific_rendering(self, viewer, data):
        viewer.add_image_layer(self.name, self.source)
        if self._contrast_controls:
            viewer.add_contrast_shader(
                self.name, black=self._black, white=self._white)


class SelectionMapper(object):
    """Class for configuring object selections based on root id

    Parameters
    ----------
    data_columns : str or list, optional
        Name (or list of names) of the data columns to get ids from. Default is None.
    fixed_ids : list, optional
        List of ids to select irrespective of data.
    """

    def __init__(self, data_columns=None, fixed_ids=None, fixed_id_colors=None, color_column=None):
        if isinstance(data_columns, str):
            data_columns = [data_columns]
        self._config = dict(data_columns=data_columns,
                            fixed_ids=fixed_ids,
                            fixed_id_colors=fixed_id_colors,
                            color_column=color_column)

    @property
    def data_columns(self):
        if self._config.get('data_columns', None) is None:
            return []
        else:
            return self._config.get('data_columns')

    @property
    def fixed_ids(self):
        if self._config.get('fixed_ids', None) is None:
            return []
        else:
            return self._config.get('fixed_ids', None)

    @property
    def fixed_id_colors(self):
        if self._config.get('fixed_id_colors', None) is None:
            return []
        else:
            return list(self._config.get('fixed_id_colors', None))

    @property
    def color_column(self):
        return self._config.get('color_column', None)

    def selected_ids(self, data):
        """ Uses the rules to generate a list of ids from a dataframe.
        """
        selected_ids = []
        if data is not None:
            for col in self.data_columns:
                selected_ids.append(data[col])
        selected_ids.append(np.array(self.fixed_ids, dtype=np.int64))
        return np.concatenate(selected_ids)

    def seg_colors(self, data):
        colors = {}
        if len(self.fixed_id_colors) == len(self.fixed_ids):
            for ii, oid in enumerate(self.fixed_ids):
                colors[oid] = self.fixed_id_colors[ii]

        if self.color_column is not None:
            clist = data[self.color_column].to_list()
            for col in self.data_columns:
                for ii, oid in enumerate(data[col]):
                    colors[oid] = clist[ii]

        return colors


class SegmentationLayerConfig(LayerConfigBase):
    """Configuration class for segmentation layers

    Parameters
    ----------
    source : str
        Segmentation source
    name : str, optional,
        Layer name.
    selected_ids_column : str or list-like, optional.
        Column name (or list of column names) to use for selected ids.
    fixed_ids : list-like, optional.
        List of root ids to select directly.
    fixed_id_colors : list-like, optional.
        List of colors for fixed ids. Should be the same length as fixed_ids, although null entries can be padded with None values.
    color_column : str, optional.
        # at the begining.
        Column to use for color values for selected objects. Values should be RGB hex strings with a
    active : bool, optional.
        If True, makes the layer selected. Default is False.
    view_kws : dict, optional.
        Keyword arguments for viewer.set_segmentation_view_options. Sets selected (and unselected) segmetation alpha values. Defaults to values in DEFAULT_SEGMENTATION_VIEW_KWS dict specified in this module.
   """

    def __init__(self,
                 source,
                 name=DEFAULT_SEG_LAYER,
                 selected_ids_column=None,
                 fixed_ids=None,
                 fixed_id_colors=None,
                 color_column=None,
                 active=False,
                 view_kws={},
                 ):
        super(SegmentationLayerConfig, self).__init__(
            name=name, type='segmentation', source=source, color=None, active=active)
        if selected_ids_column is not None or fixed_ids is not None:
            self._selection_map = SelectionMapper(
                data_columns=selected_ids_column,
                fixed_ids=fixed_ids,
                fixed_id_colors=fixed_id_colors,
                color_column=color_column)
        else:
            self._selection_map = None

        base_seg_kws = DEFAULT_SEGMENTATION_VIEW_KWS.copy()
        base_seg_kws.update(view_kws)
        self._view_kws = view_kws

    def add_selection_map(self, selected_ids_column=None, fixed_ids=None):
        if self._selection_map is not None:
            if isinstance(selected_ids_column, str):
                selected_ids_column = [selected_ids_column]
            data_columns = self._selection_map.data_columns.extend(
                selected_ids_column)
            fixed_ids = self._selection_map.fixed_ids.extend(fixed_ids)

        self._selection_map = SelectionMapper(
            data_columns=data_columns,
            fixed_ids=fixed_ids)

    def _specific_rendering(self, viewer, data):
        viewer.add_segmentation_layer(self.name, self.source)
        if self._selection_map is not None:
            selected_ids = self._selection_map.selected_ids(data)
            colors = self._selection_map.seg_colors(data)
            viewer.add_selected_objects(self.name, selected_ids, colors)
        viewer.set_segmentation_view_options(self.name, **self._view_kws)


class AnnotationLayerConfig(LayerConfigBase):
    """Configuration class for annotation layers

    Parameters
    ----------
    name : str, optional
        Layer name. By default, 'annos'
    color : str, optional
        Hex color code with an initial #. By default, None
    linked_segmentation_layer : str, optional
        Name of a linked segmentation layer for selected ids. By default, None
    mapping_rules : PointMapper, LineMapper, SphereMapper or list, optional
        One rule or a list of rules mapping data to annotations. By default, []
    array_data : bool, optional
        If True, allows simple mapping where one or more arrays are passed instead of a dataframe.
        Only allows basic annotation creation, no tags, linked segmentations, or other rich features.
    tags : list, optional
        List of tags for the layer.
    active : bool, optional
        If True, makes the layer selected. Default is True (unlike for image/segmentation layers).
    """

    def __init__(self,
                 name=DEFAULT_ANNO_LAYER,
                 color=None,
                 linked_segmentation_layer=None,
                 mapping_rules=[],
                 array_data=False,
                 tags=None,
                 active=True,
                 filter_by_segmentation=False,
                 brackets_show_segmentation=True,
                 selection_shows_segmentation=True,
                 ):
        super(AnnotationLayerConfig, self).__init__(
            name=name, type='annotation', color=color, source=None, active=active)
        self._config['linked_segmentation_layer'] = linked_segmentation_layer
        self._config['filter_by_segmentation'] = filter_by_segmentation
        self._config['selection_shows_segmentation'] = selection_shows_segmentation
        self._config['brackets_show_segmentation'] = brackets_show_segmentation
        if issubclass(type(mapping_rules), AnnotationMapperBase):
            mapping_rules = [mapping_rules]
        if array_data is True:
            if len(mapping_rules) > 1:
                raise ValueError(
                    "Only one mapping rule can be set using array data")
            for mr in mapping_rules:
                mr.array_data = array_data
        self._array_data = array_data
        self._annotation_map_rules = mapping_rules
        self._tags = tags

    @property
    def linked_segmentation_layer(self):
        return self._config.get('linked_segmentation_layer', None)

    @property
    def filter_by_segmentation(self):
        return self._config.get('filter_by_segmentation', None)

    @property
    def selection_shows_segmentation(self):
        return self._config.get('selection_shows_segmentation', None)

    @property
    def brackets_show_segmentation(self):
        return self._config.get('brackets_show_segmentation', None)

    def _specific_rendering(self, viewer, data):
        viewer.add_annotation_layer(self.name,
                                    color=self.color,
                                    linked_segmentation_layer=self.linked_segmentation_layer,
                                    filter_by_segmentation=self.filter_by_segmentation,
                                    selection_shows_segmentation=self.selection_shows_segmentation,
                                    brackets_show_segmentation=self.brackets_show_segmentation,
                                    tags=self._tags)
        annos = []
        for rule in self._annotation_map_rules:
            rule.tag_map = self._tags
            if data is not None:
                if len(data) > 0:
                    annos.extend(rule._render_data(data))
                    viewer.set_view_options(position=rule._get_position(data))
        viewer.add_annotations(self.name, annos)


class AnnotationMapperBase(object):
    def __init__(self,
                 type,
                 data_columns,
                 description_column,
                 linked_segmentation_column,
                 tag_column,
                 group_column,
                 set_position,
                 gather_linked_segmentations,
                 share_linked_segmentations,
                 ):

        self._config = dict(type=type,
                            data_columns=data_columns,
                            array_data=False,
                            description_column=description_column,
                            linked_segmentation_column=linked_segmentation_column,
                            tag_column=tag_column,
                            group_column=group_column,
                            set_position=set_position,
                            gather_linked_segmentations=gather_linked_segmentations,
                            share_linked_segmentations=share_linked_segmentations,
                            )
        self._tag_map = None

    @property
    def type(self):
        return self._config.get('type', None)

    @property
    def data_columns(self):
        return self._config.get('data_columns', None)

    @property
    def description_column(self):
        return self._config.get('description_column', None)

    @property
    def linked_segmentation_column(self):
        return self._config.get('linked_segmentation_column', None)

    @property
    def tag_column(self):
        return self._config.get('tag_column', None)

    @property
    def group_column(self):
        return self._config.get('group_column', None)

    @property
    def gather_linked_segmentations(self):
        return self._config.get('gather_linked_segmentations', True)

    @property
    def share_linked_segmentations(self):
        return self._config.get('share_linked_segmentations', False)

    @property
    def set_position(self):
        return self._config.get('set_position', False)

    @property
    def tag_map(self):
        if self._tag_map is None:
            return {}
        else:
            return self._tag_map

    @tag_map.setter
    def tag_map(self, tag_list):
        if tag_list is None:
            self._tag_map = {}
        else:
            self._tag_map = {tag: ii+1 for ii, tag in enumerate(tag_list)}

    @property
    def array_data(self):
        return self._config.get('array_data', False)

    @array_data.setter
    def array_data(self, new_array_data):
        self._config['array_data'] = new_array_data
        if new_array_data:
            self._config['data_columns'] = self._default_array_data_columns()
            self._config['description_column'] = None
            self._config['linked_segmentation_column'] = None
            self._config['tag_column'] = None

    def _default_array_data_columns(self):
        return []

    def _assign_tags(self, data):
        if self.tag_column is not None:
            anno_tags = []
            for row in data[self.tag_column]:
                if isinstance(row, Collection) and not isinstance(row, str):
                    add_annos = [self.tag_map.get(r, None) for r in row]
                else:
                    add_annos = [self.tag_map.get(row, None)]
                anno_tags.append(add_annos)
        else:
            anno_tags = [[None] for x in range(len(data))]
        return anno_tags

    def _render_data(self, data):
        # Set per subclass
        return None

    def _linked_segmentations(self, data):
        if self.linked_segmentation_column is not None:
            seg_array = np.vstack(
                data[self.linked_segmentation_column].values)
            linked_segs = [row[~pd.isnull(row)].astype(int)
                           for row in seg_array]
        else:
            linked_segs = [None for x in range(len(data))]
        return linked_segs

    def _descriptions(self, data):
        if self.description_column is not None:
            descriptions = data[self.description_column].values
        else:
            descriptions = [None for x in range(len(data))]
        return descriptions

    def _add_groups(self, groups, annos):
        vals, inverse = np.unique(groups, return_inverse=True)
        inv_inds = np.flatnonzero(~pd.isnull(vals))
        group_annos = []

        for ii in inv_inds:
            anno_to_group = [annos[jj] for jj in np.flatnonzero(inverse == ii)]
            group_annos.append(
                annotation.group_annotations(
                    anno_to_group,
                    return_all=False,
                    gather_linked_segmentations=self.gather_linked_segmentations,
                    share_linked_segmentations=self.share_linked_segmentations
                )
            )
        annos.extend(group_annos)
        return annos

    def _get_position(self, data):
        if len(data) > 0 and self.set_position is True:
            return list(data[self.data_columns[0]].iloc[0])
        else:
            return None


class PointMapper(AnnotationMapperBase):
    """Sets rules to map dataframes to point annotations

    Parameters
    ----------
    point_column : str, optional
        Column name with 3d position data
    array_data : bool, optional
        If True, the expected data is a Nx3 numpy array. This will only work for point positions,
        to use any other features you must build a dataframe.
    decription_column : str, optional
        Column name with string data for annotation descriptions
    linked_segmentation_column : str, optional
        Column name for root ids to link to annotations
    tag_column : str, optional
        Column name for categorical tag data. Tags must match those set in the
        annotation layer.
    group_column : str, optional
        Column name for grouping data. Data in this row should be numeric with possible NaNs.
        Rows with the same non-NaN value will be collected into a grouped annotation.
    set_position : bool, optional
        If set to True, moves the position to center on the first point in the
        data.
    """

    def __init__(self,
                 point_column=None,
                 description_column=None,
                 linked_segmentation_column=None,
                 tag_column=None,
                 group_column=None,
                 gather_linked_segmentations=True,
                 share_linked_segmentations=False,
                 set_position=False,
                 ):
        super(PointMapper, self).__init__(type='point',
                                          data_columns=[point_column],
                                          description_column=description_column,
                                          linked_segmentation_column=linked_segmentation_column,
                                          tag_column=tag_column,
                                          group_column=group_column,
                                          gather_linked_segmentations=gather_linked_segmentations,
                                          share_linked_segmentations=share_linked_segmentations,
                                          set_position=set_position,
                                          )

    def _default_array_data_columns(self):
        return ['pt']

    def _render_data(self, data):
        if self.array_data:
            data = pd.DataFrame(data={'pt': np.array(data).tolist()})
        col = self.data_columns[0]
        relinds = ~pd.isnull(data[col])

        pts = np.vstack(data[col][relinds])
        descriptions = self._descriptions(data[relinds])

        linked_segs = self._linked_segmentations(data[relinds])
        tags = self._assign_tags(data)
        annos = [annotation.point_annotation(pt,
                                             description=d,
                                             linked_segmentation=ls,
                                             tag_ids=t) for
                 pt, d, ls, t in zip(pts, descriptions, linked_segs, tags)]

        if self.group_column is not None:
            groups = data[self.group_column].values[relinds]
            annos = self._add_groups(groups, annos)

        return annos


class LineMapper(AnnotationMapperBase):
    """Sets rules to map dataframes to line annotations

    Parameters
    ----------
    point_column_a : str
        Column name with 3d position data for the first point of the line.
        Must be set if array_data is False (the default)
    point_column_b : str
        Column name with 3d position data for the first point of the line.
        Must be set if array_data is False (the default)
    decription_column : str, optional
        Column name with string data for annotation descriptions
    linked_segmentation_column : str, optional
        Column name for root ids to link to annotations
    tag_column : str, optional
        Column name for categorical tag data. Tags must match those set in the annotation layer.
    group_column : str, optional
        Column name for grouping data. Data in this row should be numeric with possible NaNs.
        Rows with the same non-NaN value will be collected into a grouped annotation.
    set_position : bool, optional
        If set to True, moves the position to center on the first point in the data (using point_column_a).
    """

    def __init__(self,
                 point_column_a=None,
                 point_column_b=None,
                 description_column=None,
                 linked_segmentation_column=None,
                 tag_column=None,
                 group_column=None,
                 gather_linked_segmentations=True,
                 share_linked_segmentations=False,
                 set_position=False,
                 ):
        super(LineMapper, self).__init__(type='line',
                                         data_columns=[
                                              point_column_a, point_column_b],
                                         description_column=description_column,
                                         linked_segmentation_column=linked_segmentation_column,
                                         tag_column=tag_column,
                                         group_column=group_column,
                                         gather_linked_segmentations=gather_linked_segmentations,
                                         share_linked_segmentations=share_linked_segmentations,
                                         set_position=set_position,
                                         )

    def _default_array_data_columns(self):
        return ['pt_a', 'pt_b']

    def _render_data(self, data):
        if self.array_data:
            data = pd.DataFrame(
                data={'pt_a': data[0].tolist(), 'pt_b': data[1].tolist()})
        colA, colB = self.data_columns

        relinds = np.logical_and(~pd.isnull(
            data[colA]), ~pd.isnull(data[colB]))

        ptAs = np.vstack(data[colA][relinds])
        ptBs = np.vstack(data[colB][relinds])
        descriptions = self._descriptions(data[relinds])
        linked_segs = self._linked_segmentations(data[relinds])
        tags = self._assign_tags(data)
        annos = [annotation.line_annotation(ptA, ptB,
                                            description=d,
                                            linked_segmentation=ls,
                                            tag_ids=t) for
                 ptA, ptB, d, ls, t in zip(ptAs, ptBs, descriptions, linked_segs, tags)]

        if self.group_column is not None:
            groups = data[self.group_column].values[relinds]
            annos = self._add_groups(groups, annos)

        return annos


class SphereMapper(AnnotationMapperBase):
    """Sets rules to map dataframes to sphere annotations

    Parameters
    ----------
    center_column : str
        Column name with 3d position data for the center of the sphere
    radius_column : str
        Column name with a radius for the sphere (in nm)
    decription_column : str, optional
        Column name with string data for annotation descriptions
    linked_segmentation_column : str, optional
        Column name for root ids to link to annotations
    tag_column : str, optional
        Column name for categorical tag data. Tags must match those set in the annotation layer.
    group_column : str, optional
        Column name for grouping data. Data in this row should be numeric with possible NaNs.
        Rows with the same non-NaN value will be collected into a grouped annotation.
    set_position : bool, optional
        If set to True, moves the position to center on the first point in the data.
    """

    def __init__(self,
                 center_column=None,
                 radius_column=None,
                 description_column=None,
                 linked_segmentation_column=None,
                 tag_column=None,
                 group_column=None,
                 gather_linked_segmentations=True,
                 share_linked_segmentations=False,
                 z_multiplier=0.1,
                 set_position=False,
                 ):
        super(SphereMapper, self).__init__(type='sphere',
                                           data_columns=[
                                               center_column, radius_column],
                                           description_column=description_column,
                                           linked_segmentation_column=linked_segmentation_column,
                                           tag_column=tag_column,
                                           group_column=group_column,
                                           gather_linked_segmentations=gather_linked_segmentations,
                                           share_linked_segmentations=share_linked_segmentations,
                                           set_position=set_position,
                                           )
        self._z_multiplier = z_multiplier

    def _default_array_data_columns(self):
        return ['ctr_pt', 'rad']

    def _render_data(self, data):
        if self.array_data:
            data = pd.DataFrame(
                data={'ctr_pt': data[0].tolist(), 'rad': data[1]})
        col_ctr, col_rad = self.data_columns
        relinds = np.logical_and(~pd.isnull(
            data[col_ctr]), ~pd.isnull(data[col_rad]))

        pts = np.vstack(data[col_ctr][relinds])
        rs = data[col_rad][relinds].values
        descriptions = self._descriptions(data[relinds])
        linked_segs = self._linked_segmentations(data[relinds])
        tags = self._assign_tags(data)
        annos = [annotation.sphere_annotation(pt, r,
                                              description=d,
                                              linked_segmentation=ls,
                                              tag_ids=t,
                                              z_multiplier=self._z_multiplier) for
                 pt, r, d, ls, t in zip(pts, rs, descriptions, linked_segs, tags)]

        if self.group_column is not None:
            groups = data[self.group_column].values[relinds]
            annos = self._add_groups(groups, annos)

        return annos


class StateBuilder():
    """A class for schematic mapping data frames into neuroglancer states.
    Parameters
    ----------
    """

    def __init__(
        self,
        layers=[],
        base_state=None,
        url_prefix=default_neuroglancer_base,
        state_server=None,
        resolution=[4, 4, 40],
        view_kws={},
    ):
        self._base_state = base_state
        self._layers = layers
        self._resolution = resolution
        self._url_prefix = url_prefix
        self._state_server = state_server

        base_kws = DEFAULT_VIEW_KWS.copy()
        base_kws.update(view_kws)
        self._view_kws = base_kws

    def _reset_state(self, base_state=None):
        """
        Resets the neuroglancer state status to a default viewer.
        """
        if base_state is None:
            base_state = self._base_state
        self._temp_viewer = EasyViewer()
        self._temp_viewer.set_state(base_state)

    def initialize_state(self, base_state=None):
        """Generate a new Neuroglancer state with layers as needed for the schema.

        Parameters
        ----------
        base_state : str, optional
            Optional initial state to build on, described by its JSON. By default None.
        """
        self._reset_state(base_state)

        if self._state_server is not None:
            self._temp_viewer.set_state_server(self._state_server)

        if self._resolution is not None:
            self._temp_viewer.set_resolution(self._resolution)
        self._temp_viewer.set_view_options(**self._view_kws)

    def render_state(self, data=None, base_state=None, return_as='url',
                     url_prefix=None, link_text='Neuroglancer Link'):
        """Build a Neuroglancer state out of a DataFrame.

        Parameters
        ----------
        data : pandas.DataFrame, optional
            DataFrame to use as a point source. By default None, for which
            it will return only the base_state and any fixed values.
        base_state : str, optional
            Initial state to build on, expressed as Neuroglancer JSON. By default None
        return_as : ['url', 'viewer', 'html', 'json', 'dict', 'shared'], optional
            Choice of output types. Note that if a viewer is returned, the state is not reset.
                url : Returns the raw url describing the state
                viewer : Returns an EasyViewer object holding the state information
                html : Returns an HTML link to the url, useful for notebooks.
                json : Returns a JSON string describing the state.
                dict : Returns a dict version of the JSON state.
            By default 'url'
        url_prefix : str, optional
            Neuroglancer URL prefix to use. By default None, for which it will open with the
            class default.
        link_text : str, optional
            Text to use for the link when returning as html, by default 'Neuroglancer Link'
        Returns
        -------
        string or neuroglancer.Viewer
            A link to or viewer for a Neuroglancer state with layers, annotations, and selected objects determined by the data.
        """
        if base_state is None:
            base_state = self._base_state
        self.initialize_state(base_state=base_state)

        if url_prefix is None:
            url_prefix = self._url_prefix

        self._render_layers(data)

        if return_as == 'viewer':
            return self.viewer
        elif return_as == 'url':
            url = self._temp_viewer.as_url(prefix=url_prefix)
            self.initialize_state()
            return url
        elif return_as == 'html':
            out = self._temp_viewer.as_url(
                prefix=url_prefix, as_html=True, link_text=link_text)
            out = HTML(out)
            self.initialize_state()
            return out
        elif return_as == 'dict':
            out = self._temp_viewer.state.to_json()
            self.initialize_state()
            return out
        elif return_as == 'json':
            from json import dumps
            out = self._temp_viewer.state.to_json()
            self.initialize_state()
            return dumps(out)
        else:
            raise ValueError('No appropriate return type selected')

    def _render_layers(self, data):
        for layer in self._layers:
            layer._render_layer(self._temp_viewer, data)

    @property
    def viewer(self):
        return self._temp_viewer


class ChainedStateBuilder():
    def __init__(self, statebuilders):
        """Builds a collection of states that sequentially add annotations based on a sequence of dataframes.

        Parameters
        ----------
        statebuilders : list
            List of DataStateBuilders, in same order as dataframes will be passed
        """
        self._statebuilders = statebuilders
        if len(self._statebuilders) == 0:
            raise ValueError('Must have at least one statebuilder')

    def render_state(self, data_list=None, base_state=None, return_as='url',
                     url_prefix=default_neuroglancer_base, link_text='Neuroglancer Link'):
        """Generate a single neuroglancer state by addatively applying an ordered collection of
        dataframes to an collection of StateBuilder renders.
        Parameters
            data_list : Collection of DataFrame. The order must match the order of StateBuilders
                        contained in the class on creation.
            base_state : JSON neuroglancer state (optional, default is None).
                         Used as a base state for adding everything else to.
            return_as: ['url', 'viewer', 'html', 'json']. optional, default='url'.
                       Sets how the state is returned. Note that if a viewer is returned,
                       the state is not reset to default.
            url_prefix: string, optional (default is https://neuromancer-seung-import.appspot.com).
                        Overrides the default neuroglancer url for url generation.
        """
        if data_list is None:
            data_list = len(self._statebuilders) * [None]

        if len(data_list) != len(self._statebuilders):
            raise ValueError('Must have as many dataframes as statebuilders')

        temp_state = base_state
        for builder, data in zip(self._statebuilders[:-1], data_list[:-1]):
            temp_state = builder.render_state(data=data,
                                              base_state=temp_state,
                                              return_as='dict')
        last_builder = self._statebuilders[-1]
        return last_builder.render_state(data=data_list[-1],
                                         base_state=temp_state,
                                         return_as=return_as,
                                         url_prefix=url_prefix)



from .statebuilder import StateBuilder, \
    ChainedStateBuilder, \
    SelectionMapper, \
    SegmentationLayerConfig, \
    AnnotationLayerConfig, \
    ImageLayerConfig, \
    PointMapper, \
    LineMapper, \
    SphereMapper



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for representing the Neuroglancer viewer state."""

from __future__ import absolute_import

import collections
import copy
import math

import numpy as np
import six

from .equivalence_map import EquivalenceMap
from .json_utils import encode_json_for_repr
from .json_wrappers import (
    JsonObjectWrapper,
    array_wrapper,
    optional,
    text_type,
    typed_list,
    typed_set,
    typed_string_map,
    wrapped_property,
)

__all__ = []


def export(obj):
    __all__.append(obj.__name__)
    return obj


def interpolate_linear(a, b, t):
    return a * (1 - t) + b * t


@export
class SpatialPosition(JsonObjectWrapper):
    __slots__ = ()
    voxel_size = voxelSize = wrapped_property(
        "voxelSize", optional(array_wrapper(np.float32, 3))
    )
    spatial_coordinates = spatialCoordinates = wrapped_property(
        "spatialCoordinates", optional(array_wrapper(np.float32, 3))
    )
    voxel_coordinates = voxelCoordinates = wrapped_property(
        "voxelCoordinates", optional(array_wrapper(np.float32, 3))
    )

    @staticmethod
    def interpolate(a, b, t):
        if (
            a.voxel_size is None
            or a.voxel_coordinates is None
            or b.voxel_coordinates is None
        ):
            return a
        c = copy.deepcopy(a)
        c.voxel_coordinates = interpolate_linear(
            a.voxel_coordinates, b.voxel_coordinates, t
        )
        return c


def unit_quaternion():
    return np.array([0, 0, 0, 1], np.float32)


def quaternion_slerp(a, b, t):
    """Spherical linear interpolation for unit quaternions.

    This is based on the implementation in the gl-matrix package:
    https://github.com/toji/gl-matrix
    """
    if a is None:
        a = unit_quaternion()
    if b is None:
        b = unit_quaternion()
    # calc cosine
    cosom = np.dot(a, b)
    # adjust signs (if necessary)
    if cosom < 0.0:
        cosom = -cosom
        b = -b

    # calculate coefficients
    if (1.0 - cosom) > 0.000001:
        # standard case (slerp)
        omega = math.acos(cosom)
        sinom = math.sin(omega)
        scale0 = math.sin((1.0 - t) * omega) / sinom
        scale1 = math.sin(t * omega) / sinom
    else:
        # "from" and "to" quaternions are very close
        #  ... so we can do a linear interpolation
        scale0 = 1.0 - t
        scale1 = t
    return scale0 * a + scale1 * b


@export
class Pose(JsonObjectWrapper):
    __slots__ = ()
    position = wrapped_property("position", SpatialPosition)
    orientation = wrapped_property(
        "orientation", optional(array_wrapper(np.float32, 4))
    )

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.position = SpatialPosition.interpolate(a.position, b.position, t)
        c.orientation = quaternion_slerp(a.orientation, b.orientation, t)
        return c


def interpolate_zoom(a, b, t):
    if a is None or b is None:
        return a
    scale_change = math.log(b / a)
    return a * math.exp(scale_change * t)


@export
class NavigationState(JsonObjectWrapper):
    __slots__ = ()
    pose = wrapped_property("pose", Pose)
    zoom_factor = zoomFactor = wrapped_property("zoomFactor", optional(float))

    @property
    def position(self):
        return self.pose.position

    @position.setter
    def position(self, v):
        self.pose.position = v

    @property
    def voxel_size(self):
        return self.pose.position.voxel_size

    @voxel_size.setter
    def voxel_size(self, v):
        self.pose.position.voxel_size = v

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.pose = Pose.interpolate(a.pose, b.pose, t)
        c.zoom_factor = interpolate_zoom(a.zoom_factor, b.zoom_factor, t)
        return c


@export
class Layer(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", optional(text_type))
    tab = wrapped_property("tab", optional(text_type))


@export
class PointAnnotationLayer(Layer):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PointAnnotationLayer, self).__init__(
            *args, type="pointAnnotation", **kwargs
        )

    points = wrapped_property(
        "points", typed_list(array_wrapper(np.float32, 3)))


def volume_source(x):
    if not isinstance(x, six.string_types):
        raise TypeError
    return text_type(x)


class _AnnotationLayerOptions(object):
    __slots__ = ()
    annotation_color = annotationColor = wrapped_property(
        "annotationColor", optional(text_type)
    )
    annotation_fill_opacity = annotationFillOpacity = wrapped_property(
        "annotationFillOpacity", optional(float, 0)
    )


@export
class ImageLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(ImageLayer, self).__init__(*args, type="image", **kwargs)

    source = wrapped_property("source", volume_source)
    shader = wrapped_property("shader", text_type)
    opacity = wrapped_property("opacity", optional(float, 0.5))
    blend = wrapped_property("blend", optional(str))
    cross_section_render_scale = crossSectionRenderScale = wrapped_property(
        "crossSectionRenderScale", optional(float, 1)
    )

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.opacity = interpolate_linear(a.opacity, b.opacity, t)
        return c


def uint64_equivalence_map(obj, _readonly=False):
    if isinstance(obj, EquivalenceMap):
        return obj
    if obj is not None:
        obj = [[int(v) for v in group] for group in obj]
    return EquivalenceMap(obj, _readonly=_readonly)


def annotation(obj, _readonly=False):
    if isinstance(obj, AnnotationBase):
        obj = obj.to_json()
    elif not isinstance(obj, dict):
        raise TypeError
    t = obj.get("type")
    return annotation_types[t](obj, _readonly=_readonly)


@export
class SegmentationLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(SegmentationLayer, self).__init__(
            *args, type="segmentation", **kwargs)

    source = wrapped_property("source", optional(volume_source))
    mesh = wrapped_property("mesh", optional(text_type))
    skeletons = wrapped_property("skeletons", optional(text_type))
    segments = wrapped_property("segments", typed_set(np.uint64))
    equivalences = wrapped_property("equivalences", uint64_equivalence_map)
    hide_segment_zero = hideSegmentZero = wrapped_property(
        "hideSegmentZero", optional(bool, True)
    )
    selected_alpha = selectedAlpha = wrapped_property(
        "selectedAlpha", optional(float, 0.5)
    )
    not_selected_alpha = notSelectedAlpha = wrapped_property(
        "notSelectedAlpha", optional(float, 0)
    )
    object_alpha = objectAlpha = wrapped_property(
        "objectAlpha", optional(float, 1.0))
    skeleton_shader = skeletonShader = wrapped_property(
        "skeletonShader", text_type)
    color_seed = colorSeed = wrapped_property("colorSeed", optional(int, 0))
    cross_section_render_scale = crossSectionRenderScale = wrapped_property(
        "crossSectionRenderScale", optional(float, 1)
    )
    mesh_render_scale = meshRenderScale = wrapped_property(
        "meshRenderScale", optional(float, 10)
    )

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for k in ["selected_alpha", "not_selected_alpha", "object_alpha"]:
            setattr(c, k, interpolate_linear(getattr(a, k), getattr(b, k), t))
        return c


@export
class annotationHolder(JsonObjectWrapper):
    __slots__ = ()
    annotations = wrapped_property("annotations", typed_list(annotation))
    tags = wrapped_property("tags", optional(list))


@export
class ChunkedgraphSegmentationLayer(SegmentationLayer):
    def __init__(self, *args, **kwargs):
        super(SegmentationLayer, self).__init__(
            *args, type="segmentation_with_graph", **kwargs
        )

    graph_operation_marker = graphOperationMarker = wrapped_property(
        "graphOperationMarker", typed_list(annotationHolder)
    )


@export
class SingleMeshLayer(Layer):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(SingleMeshLayer, self).__init__(*args, type="mesh", **kwargs)

    source = wrapped_property("source", text_type)
    vertex_attribute_sources = vertexAttributeSources = wrapped_property(
        "vertexAttributeSources", optional(typed_list(text_type))
    )
    shader = wrapped_property("shader", text_type)
    vertex_attribute_names = vertexAttributeNames = wrapped_property(
        "vertexAttributeNames", optional(typed_list(optional(text_type)))
    )


class AnnotationBase(JsonObjectWrapper):
    __slots__ = ()

    id = wrapped_property("id", optional(text_type)
                          )  # pylint: disable=invalid-name
    type = wrapped_property("type", text_type)
    description = wrapped_property("description", optional(text_type))
    segments = wrapped_property("segments", optional(typed_list(np.uint64)))
    parent_id = wrapped_property("parentId", optional(text_type))
    tag_ids = wrapped_property("tagIds", optional(list))


@export
class PointAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(PointAnnotation, self).__init__(*args, type="point", **kwargs)

    point = wrapped_property("point", array_wrapper(np.float32, 3))


@export
class LineAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(LineAnnotation, self).__init__(*args, type="line", **kwargs)

    point_a = pointA = wrapped_property("pointA", array_wrapper(np.float32, 3))
    point_b = pointB = wrapped_property("pointB", array_wrapper(np.float32, 3))


@export
class AxisAlignedBoundingBoxAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(AxisAlignedBoundingBoxAnnotation, self).__init__(
            *args, type="axis_aligned_bounding_box", **kwargs
        )

    point_a = pointA = wrapped_property("pointA", array_wrapper(np.float32, 3))
    point_b = pointB = wrapped_property("pointB", array_wrapper(np.float32, 3))


@export
class EllipsoidAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(EllipsoidAnnotation, self).__init__(
            *args, type="ellipsoid", **kwargs)

    center = wrapped_property("center", array_wrapper(np.float32, 3))
    radii = wrapped_property("radii", array_wrapper(np.float32, 3))


@export
class CollectionAnnotation(AnnotationBase):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(CollectionAnnotation, self).__init__(
            *args, type="collection", **kwargs)

    source = wrapped_property("source", array_wrapper(np.float32, 3))
    entries = wrapped_property("entries", array_wrapper(list))
    children_visible = wrapped_property("childrenVisible", array_wrapper(bool))


annotation_types = {
    "point": PointAnnotation,
    "line": LineAnnotation,
    "axis_aligned_bounding_box": AxisAlignedBoundingBoxAnnotation,
    "ellipsoid": EllipsoidAnnotation,
    "collection": CollectionAnnotation,
}


annotation.supports_readonly = True


@export
class AnnotationLayer(Layer, _AnnotationLayerOptions):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super(AnnotationLayer, self).__init__(
            *args, type="annotation", **kwargs)

    source = wrapped_property("source", optional(volume_source))
    voxel_size = voxelSize = wrapped_property(
        "voxelSize", optional(array_wrapper(np.float32, 3))
    )
    annotations = wrapped_property("annotations", typed_list(annotation))
    linked_segmentation_layer = linkedSegmentationLayer = wrapped_property(
        "linkedSegmentationLayer", optional(text_type)
    )
    filter_by_segmentation = filterBySegmentation = wrapped_property(
        "filterBySegmentation", optional(bool, False)
    )
    brackets_show_segmentation = bracketShowSegmentation = wrapped_property(
        "bracketShortcutsShowSegmentation", optional(bool, True)
    )
    selection_shows_segmentation = annotationSelectionShowsSegmentation = wrapped_property(
        "annotationSelectionShowsSegmentation", optional(bool, True)
    )

    @staticmethod
    def interpolate(a, b, t):
        del b
        del t
        return a


layer_types = {
    "image": ImageLayer,
    "segmentation": SegmentationLayer,
    "segmentation_with_graph": ChunkedgraphSegmentationLayer,
    "pointAnnotation": PointAnnotationLayer,
    "annotation": AnnotationLayer,
    "mesh": SingleMeshLayer,
}


def make_layer(json_data, _readonly=False):
    if isinstance(json_data, Layer):
        return json_data

    if not isinstance(json_data, dict):
        raise TypeError

    type_name = json_data.get("type")
    layer_type = layer_types.get(type_name)
    if layer_type is not None:
        return layer_type(json_data, _readonly=_readonly)
    else:
        raise ValueError


@export
class ManagedLayer(JsonObjectWrapper):
    __slots__ = ("name", "layer")

    def __init__(self, name, layer=None, _readonly=False, **kwargs):
        if isinstance(name, ManagedLayer):
            if layer is not None or kwargs:
                raise ValueError
            layer = name.to_json()
            name = name.name

        object.__setattr__(self, "name", name)

        if isinstance(layer, Layer):
            json_data = collections.OrderedDict()
        else:
            if layer is None:
                json_data = collections.OrderedDict()
            else:
                json_data = layer
            layer = make_layer(json_data, _readonly=_readonly)

        object.__setattr__(self, "layer", layer)
        super(ManagedLayer, self).__init__(
            json_data, _readonly=_readonly, **kwargs)

    visible = wrapped_property("visible", optional(bool))

    def __getattr__(self, key):
        return getattr(self.layer, key)

    def __setattr__(self, key, value):
        if self._readonly:
            raise AttributeError
        if key in ["name", "visible", "layer"]:
            object.__setattr__(self, key, value)
        else:
            return setattr(self.layer, key, value)

    def __repr__(self):
        return u"ManagedLayer(%s,%s)" % (
            encode_json_for_repr(self.name),
            encode_json_for_repr(self.to_json()),
        )

    def to_json(self):
        r = self.layer.to_json()
        r["name"] = self.name
        visible = self.visible
        if visible is not None:
            r["visible"] = visible
        return r

    def __deepcopy__(self, memo):
        return ManagedLayer(self.name, copy.deepcopy(self.to_json(), memo))


@export
class Layers(object):
    __slots__ = ("_layers", "_readonly")
    supports_readonly = True

    def __init__(self, json_data, _readonly=False):
        if json_data is None:
            json_data = collections.OrderedDict()
        self._layers = []
        self._readonly = _readonly
        if isinstance(json_data, collections.abc.Mapping):
            for k, v in six.iteritems(json_data):
                self._layers.append(ManagedLayer(k, v, _readonly=_readonly))
        else:
            # layers property can also be an array in JSON now. each layer has a name property
            for layer in json_data:
                if isinstance(layer, ManagedLayer):
                    self._layers.append(
                        ManagedLayer(layer.name, layer, _readonly=_readonly)
                    )
                elif isinstance(layer, dict):
                    self._layers.append(
                        ManagedLayer(
                            text_type(layer["name"]), layer, _readonly=_readonly
                        )
                    )
                else:
                    raise TypeError

    def index(self, k):
        for i, u in enumerate(self._layers):
            if u.name == k:
                return i
        return -1

    def __getitem__(self, k):
        """Indexes into the list of layers by index, slice, or layer name."""
        if isinstance(k, six.string_types):
            return self._layers[self.index(k)]
        return self._layers[k]

    def __setitem__(self, k, v):
        if self._readonly:
            raise AttributeError
        if isinstance(k, six.string_types):
            i = self.index(k)
            if isinstance(v, Layer):
                v = ManagedLayer(k, v)
            elif not isinstance(v, ManagedLayer):
                raise TypeError
            if i == -1:
                self._layers.append(v)
            else:
                self._layers[i] = v
        else:
            if isinstance(k, slice):
                values = []
                for x in v:
                    if not isinstance(v, ManagedLayer):
                        raise TypeError
                    values.append(x)
                self._layers[k] = values
            else:
                if not isinstance(v, ManagedLayer):
                    raise TypeError
                self._layers[k] = v

    def clear(self):
        """Clears the list of layers."""
        del self[:]

    def __delitem__(self, k):
        """Deletes a layer by index, slice, or name."""
        if self._readonly:
            raise AttributeError
        if isinstance(k, six.string_types):
            k = self.index(k)
        del self._layers[k]

    def append(self, *args, **kwargs):
        """Appends a ManagedLayer to the list of layers."""
        if self._readonly:
            raise AttributeError
        if len(args) == 1 and not kwargs and isinstance(args[0], ManagedLayer):
            layer = args[0]
        else:
            layer = ManagedLayer(*args, **kwargs)
        self._layers.append(layer)

    def extend(self, elements):
        for element in elements:
            self.append(element)

    def __len__(self):
        """Returns the number of layers in the list."""
        return len(self._layers)

    def __iter__(self):
        return iter(self._layers)

    def to_json(self):
        r = []
        for x in self._layers:
            r.append(x.to_json())
        return r

    def __repr__(self):
        return repr(self._layers)

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for layer in c:
            index = b.index(layer.name)
            if index == -1:
                continue
            other_layer = b[index]
            if type(other_layer.layer) is not type(
                layer.layer
            ):  # pylint: disable=unidiomatic-typecheck
                continue
            layer.layer = type(layer.layer).interpolate(
                layer.layer, other_layer.layer, t
            )
        return c


def navigation_link_type(x):
    x = six.text_type(x)
    x = x.lower()
    if x not in [u"linked", u"unlinked", u"relative"]:
        raise ValueError("Invalid navigation link type: %r" % x)
    return x


def make_linked_navigation_type(value_type, interpolate_function=None):
    if interpolate_function is None:
        interpolate_function = value_type.interpolate

    class LinkedType(JsonObjectWrapper):
        __slots__ = ()
        link = wrapped_property("link", optional(
            navigation_link_type, u"linked"))
        value = wrapped_property("value", optional(value_type))

        @staticmethod
        def interpolate(a, b, t):
            c = copy.deepcopy(a)
            c.link = a.link
            if a.link == b.link and a.link != "linked":
                c.value = interpolate_function(a, b, t)
                return c
            return c

    return LinkedType


@export
class LinkedSpatialPosition(make_linked_navigation_type(SpatialPosition)):
    __slots__ = ()


@export
class LinkedZoomFactor(make_linked_navigation_type(float, interpolate_zoom)):
    __slots__ = ()


@export
class LinkedOrientationState(
    make_linked_navigation_type(array_wrapper(np.float32, 4), quaternion_slerp)
):
    __slots__ = ()


@export
class CrossSection(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    width = wrapped_property("width", optional(int, 1000))
    height = wrapped_property("height", optional(int, 1000))
    position = wrapped_property("position", LinkedSpatialPosition)
    orientation = wrapped_property("orientation", LinkedOrientationState)
    zoom = wrapped_property("zoom", LinkedZoomFactor)

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.width = interpolate_linear(a.width, b.width, t)
        c.height = interpolate_linear(a.height, b.height, t)
        c.position = LinkedSpatialPosition.interpolate(
            a.position, b.position, t)
        c.orientation = LinkedOrientationState.interpolate(
            a.orientation, b.orientation, t
        )
        c.zoom = LinkedZoomFactor.interpolate(a.zoom, b.zoom, t)
        return c


@export
class CrossSectionMap(typed_string_map(CrossSection)):
    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for k in a:
            if k in b:
                c[k] = CrossSection.interpolate(a[k], b[k], t)
        return c


@export
class DataPanelLayout(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", text_type)
    cross_sections = crossSections = wrapped_property(
        "crossSections", CrossSectionMap)
    orthographic_projection = orthographicProjection = wrapped_property(
        "orthographicProjection", optional(bool, False)
    )

    def __init__(self, json_data=None, _readonly=False, **kwargs):
        if isinstance(json_data, six.string_types):
            json_data = {"type": six.text_type(json_data)}
        super(DataPanelLayout, self).__init__(
            json_data, _readonly=_readonly, **kwargs)

    def to_json(self):
        if len(self.cross_sections) == 0 and not self.orthographic_projection:
            return self.type
        return super(DataPanelLayout, self).to_json()

    @staticmethod
    def interpolate(a, b, t):
        if a.type != b.type or len(a.cross_sections) == 0:
            return a
        c = copy.deepcopy(a)
        c.cross_sections = CrossSectionMap.interpolate(
            a.cross_sections, b.cross_sections, t
        )
        return c


def data_panel_layout_wrapper(default_value="xy"):
    def wrapper(x, _readonly=False):
        if x is None:
            x = default_value
        if isinstance(x, six.string_types):
            x = {"type": six.text_type(x)}
        return DataPanelLayout(x, _readonly=_readonly)

    wrapper.supports_readonly = True
    return wrapper


data_panel_layout_types = frozenset(
    ["xy", "yz", "yz", "xy-3d", "yz-3d", "yz-3d", "4panel", "3d"]
)


def layout_specification(x, _readonly=False):
    if x is None:
        x = "4panel"
    if isinstance(x, six.string_types):
        x = {"type": six.text_type(x)}
    if isinstance(x, (StackLayout, LayerGroupViewer, DataPanelLayout)):
        return type(x)(x.to_json(), _readonly=_readonly)
    if not isinstance(x, dict):
        raise ValueError
    layout_type = layout_types.get(x.get("type"))
    if layout_type is None:
        raise ValueError
    return layout_type(x, _readonly=_readonly)


layout_specification.supports_readonly = True


@export
class StackLayout(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", text_type)
    children = wrapped_property("children", typed_list(layout_specification))

    def __getitem__(self, key):
        return self.children[key]

    def __len__(self):
        return len(self.children)

    def __setitem__(self, key, value):
        self.children[key] = value

    def __delitem__(self, key):
        del self.children[key]

    def __iter__(self):
        return iter(self.children)

    @staticmethod
    def interpolate(a, b, t):
        if a.type != b.type or len(a.children) != len(b.children):
            return a
        c = copy.deepcopy(a)
        c.children = [
            interpolate_layout(a_child, b_child, t)
            for a_child, b_child in zip(a.children, b.children)
        ]
        return c


@export
def row_layout(children):
    return StackLayout(type="row", children=children)


@export
def column_layout(children):
    return StackLayout(type="column", children=children)


def interpolate_layout(a, b, t):
    if type(a) is not type(b):
        return a
    return type(a).interpolate(a, b, t)


@export
class LayerGroupViewer(JsonObjectWrapper):
    __slots__ = ()
    type = wrapped_property("type", text_type)
    layers = wrapped_property("layers", typed_list(text_type))
    layout = wrapped_property("layout", data_panel_layout_wrapper("xy"))
    position = wrapped_property("position", LinkedSpatialPosition)
    cross_section_orientation = crossSectionOrientation = wrapped_property(
        "crossSectionOrientation", LinkedOrientationState
    )
    cross_section_zoom = crossSectionZoom = wrapped_property(
        "crossSectionZoom", LinkedZoomFactor
    )
    perspective_orientation = perspectiveOrientation = wrapped_property(
        "perspectiveOrientation", LinkedOrientationState
    )
    perspective_zoom = perspectiveZoom = wrapped_property(
        "perspectiveZoom", LinkedZoomFactor
    )

    def __init__(self, *args, **kwargs):
        super(LayerGroupViewer, self).__init__(*args, **kwargs)
        self.type = "viewer"

    def __repr__(self):
        j = self.to_json()
        j.pop("type", None)
        return u"%s(%s)" % (type(self).__name__, encode_json_for_repr(j))

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        for k in (
            "layout",
            "position",
            "cross_section_orientation",
            "cross_section_zoom",
            "perspective_orientation",
            "perspective_zoom",
        ):
            a_attr = getattr(a, k)
            b_attr = getattr(b, k)
            setattr(c, k, type(a_attr).interpolate(a_attr, b_attr, t))
        return c


layout_types = {
    "row": StackLayout,
    "column": StackLayout,
    "viewer": LayerGroupViewer,
}


def add_data_panel_layout_types():
    for k in data_panel_layout_types:
        layout_types[k] = DataPanelLayout


add_data_panel_layout_types()


@export
class SelectedLayerState(JsonObjectWrapper):
    visible = wrapped_property("visible", optional(bool, False))
    size = wrapped_property("size", optional(int))
    layer = wrapped_property("layer", optional(text_type))


@export
class StatisticsDisplayState(JsonObjectWrapper):
    visible = wrapped_property("visible", optional(bool, False))
    size = wrapped_property("size", optional(int))


@export
class ViewerState(JsonObjectWrapper):
    __slots__ = ()
    navigation = wrapped_property("navigation", NavigationState)
    perspective_zoom = perspectiveZoom = wrapped_property(
        "perspectiveZoom", optional(float)
    )
    perspective_orientation = perspectiveOrientation = wrapped_property(
        "perspectiveOrientation", optional(array_wrapper(np.float32, 4))
    )
    show_slices = showSlices = wrapped_property(
        "showSlices", optional(bool, True))
    show_axis_lines = showAxisLines = wrapped_property(
        "showAxisLines", optional(bool, True)
    )
    show_scale_bar = showScaleBar = wrapped_property(
        "showScaleBar", optional(bool, True)
    )
    show_default_annotations = showDefaultAnnotations = wrapped_property(
        "showDefaultAnnotations", optional(bool, True)
    )
    gpu_memory_limit = gpuMemoryLimit = wrapped_property(
        "gpuMemoryLimit", optional(int)
    )
    system_memory_limit = systemMemoryLimit = wrapped_property(
        "systemMemoryLimit", optional(int)
    )
    concurrent_downloads = concurrentDownloads = wrapped_property(
        "concurrentDownloads", optional(int)
    )
    layers = wrapped_property("layers", Layers)
    layout = wrapped_property("layout", layout_specification)
    cross_section_background_color = crossSectionBackgroundColor = wrapped_property(
        "crossSectionBackgroundColor", optional(text_type)
    )
    perspective_view_background_color = (
        perspectiveViewBackgroundColor
    ) = wrapped_property("perspectiveViewBackgroundColor", optional(text_type))
    selected_layer = selectedLayer = wrapped_property(
        "selectedLayer", SelectedLayerState
    )
    statistics = wrapped_property("statistics", StatisticsDisplayState)

    @property
    def position(self):
        return self.navigation.position

    @position.setter
    def position(self, v):
        self.navigation.position = v

    @property
    def voxel_coordinates(self):
        return self.position.voxel_coordinates

    @voxel_coordinates.setter
    def voxel_coordinates(self, v):
        self.position.voxel_coordinates = v

    @property
    def voxel_size(self):
        return self.navigation.voxel_size

    @voxel_size.setter
    def voxel_size(self, v):
        self.navigation.voxel_size = v

    @staticmethod
    def interpolate(a, b, t):
        c = copy.deepcopy(a)
        c.navigation = NavigationState.interpolate(
            a.navigation, b.navigation, t)
        c.perspective_zoom = interpolate_zoom(
            a.perspective_zoom, b.perspective_zoom, t)
        c.perspective_orientation = quaternion_slerp(
            a.perspective_orientation, b.perspective_orientation, t
        )
        c.layers = Layers.interpolate(a.layers, b.layers, t)
        c.layout = interpolate_layout(a.layout, b.layout, t)
        return c
